using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Text.RegularExpressions;
using FitConnect.Encryption;
using FitConnect.Interfaces;
using FitConnect.Models;
using FitConnect.RestService;
using FitConnect.Services;
using FitConnect.Services.Models;
using FitConnect.Services.Models.v1;
using Microsoft.Extensions.Logging;

namespace FitConnect;

/// <summary>
///     The fluent implementation for the <see cref="Sender" /> to encapsulate the FitConnect API.
/// </summary>
/// <example>
///     Reference for FluentSender
///     <code>
/// client.Sender
///     .Authenticate(clientId!, clientSecret!)
///     .CreateSubmission(new Submission { Attachments = new List() })
///     .UploadAttachments()
///     .SendSubmission(new Metadata(), new Data());
/// </code>
/// </example>
public class FluentSender : Sender, IFluentSender, IFluentSenderWithDestination,
    IFluentSenderWithAttachments, IFluentSenderWithData, IFluentSenderWithService {
    private State _nextStep = State.Authenticate;
    private OAuthAccessToken? token;

    public FluentSender(FitConnectEndpoints endpoints,
        IOAuthService oAuthService,
        ISubmissionService submissionService,
        IRouteService routeService,
        IDestinationService destinationService,
        ILogger? logger = null) : base(endpoints, oAuthService, submissionService,
        routeService, destinationService, logger) {
    }

    public string? PublicKey { get; set; }
    public CreateSubmissionDto? NewSubmission { get; set; }
    public Submission? Submission { get; set; }


    public IFluentSenderWithDestination FindDestinationId(string leiaKey, string? ags, string? ars,
        string? areaId) {
        var destinationId = RouteService.GetDestinationIdAsync(leiaKey, ags, ars, areaId).Result;
        return WithDestination(destinationId);
    }

    public IFluentSenderWithDestination FindDestinationId(Destination destination) {
        throw new NotImplementedException();
    }

    public IFluentSenderWithDestination WithDestination(string destinationId) {
        if (!Regex.IsMatch(destinationId, ".*"))
            throw new AggregateException("Invalid destination id");

        Submission = CreateSubmission(destinationId).Result;
        return this;
    }

    public IFluentSenderWithService WithService(string serviceName, string leikaKey) {
        throw new NotImplementedException();
    }

    public IFluentSenderWithData WithData(string data) {
        Submission.Data = Data.FromString(data);
        return this;
    }

    public Submission Submit() {
        throw new NotImplementedException();
    }


    /// <summary>
    ///     Authenticates with the FitConnect API.
    /// </summary>
    /// <param name="clientId">The client id from the submission service</param>
    /// <param name="clientSecret">The client secret</param>
    /// <param name="scope">Optional scope for the credentials</param>
    /// <returns></returns>
    private FluentSender
        Authenticate(string clientId, string clientSecret, string? scope = null) {
        try {
            if (_nextStep != State.Authenticate)
                throw new InvalidOperationException("Client not ready for authentication.");

            token ??= AuthenticateAsync(clientId, clientSecret, scope).Result;
            if (token == null)
                throw new AuthenticationException(
                    "Failed to authenticate with FitConnect. Token is null.");

            _nextStep = State.CreateSubmission;
            return this;
        }
        catch (Exception e) {
            throw new AuthenticationException("Failed to authenticate with FitConnect", e);
        }
    }

    /// <summary>
    ///     Creates a new <see cref="Submission" /> on the FitConnect server.
    /// </summary>
    /// <param name="destinationId">The id of the destination the submission has to be sent to</param>
    /// <returns></returns>
    /// <exception cref="InvalidOperationException">If sender is not authenticated</exception>
    /// <exception cref="ArgumentException">If submission is not ready to be sent</exception>
    private async Task<Submission> CreateSubmission(string destinationId) {
        Authenticate(Owner.ClientId, Owner.ClientSecret);
        PublicKey = await GetPublicKeyFromDestination(destinationId);
        Encryption = new FitEncryption(Logger) { PublicKeyEncryption = PublicKey };

        var submission = new Submission() {
            DestinationId = destinationId
        };

        _nextStep = State.UploadAttachments;
        return submission;
    }

    private void IntroduceSubmission(Submission submission) {
        if (submission.ServiceType == null)
            throw new ArgumentException("Submission has no service type.");

        SubmissionService.CreateSubmission((CreateSubmissionDto)submission);
    }

    private async Task<string> GetPublicKeyFromDestination(string destinationId) {
        DestinationService.Token = token!.AccessToken;
        SubmissionService.Token = token!.AccessToken;
        var publicKey = await DestinationService.GetPublicKey(destinationId);
        return publicKey;
    }


    /// <summary>
    /// 
    /// </summary>
    /// <param name="attachments"></param>
    /// <returns></returns>
    /// <exception cref="InvalidOperationException"></exception>
    /// <exception cref="ArgumentException"></exception>
    public IFluentSenderWithAttachments WithAttachments(IEnumerable<Attachment> attachments) {
        Submission!.Attachments = attachments.ToList();
        if (Submission.ServiceType == null)
            throw new ArgumentException("Submission has no service type.");

        var created = SubmissionService.CreateSubmission((CreateSubmissionDto)Submission);
        Submission.Id = created.SubmissionId;
        Submission.CaseId = created.CaseId;

        var encryptedAttachments = Encrypt(PublicKey!, Submission.Attachments);
        UploadAttachmentsAsync(Submission.Id!, encryptedAttachments).Wait();

        return this;
    }


    private enum State {
        Authenticate,
        CreateSubmission,
        UploadAttachments,
        SendSubmission,
        Complete
    }

    IFluentSender IFluentSenderReady.Submit() {
        Logger?.LogInformation("Sending submission");
        return this;
    }

    public IFluentSenderWithService WithServiceType(string serviceName, string leikaKey) {
        if (!Regex.IsMatch(leikaKey, "^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$"))
            throw new ArgumentException("Invalid leika key");

        Submission.ServiceType = new ServiceType() {
            Name = serviceName,
            Identifier = leikaKey
        };
        return this;
    }
}
